/* Doors for Linux.
   Copyright (C) 1998 Jason Lango

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
**  doorsk.c - Main kernel code for Linux/Doors.
**  Jason Lango <jal@cs.brown.edu>
*/

#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>
#include "doorsk.h"
#include "door_ioctl.h"
#include "vm_copy.h"
#include "doors_debug.h"

const char *doors_banner = DOORS_BANNER;

#define DOOR_DEFAULT_MODE 0600

#define DOOR_DESC_ALIGN_SIZE 8UL
#define DOOR_DESC_ALIGN_MASK ~(DOOR_DESC_ALIGN_SIZE-1)
#define DOOR_DESC_ALIGN(addr) \
	(((addr)+DOOR_DESC_ALIGN_SIZE-1)&DOOR_DESC_ALIGN_MASK)

static int door_open(struct inode *ino, struct file *filp);
static int door_release(struct inode *ino, struct file *filp);
static void door_put_inode(struct inode *ino);
static void door_delete_inode(struct inode *ino);

struct door_server {
	struct list_head	ds_servers;	/* link on door_servers */
	struct fs_struct	*ds_proc_ident;
	pid_t			ds_pid;
	struct wait_queue	*ds_wait_svr_create;
	struct wait_queue	*ds_wait_svr_create_ready;
	struct wait_queue	*ds_wait_avail_servers;
	struct wait_queue	*ds_wait_server_ready;
	int			ds_needs_svr_create;
	struct list_head	ds_doors;
	struct list_head	ds_inprogress;	/* messages in progress */
};
#define task_pid_cookie(task) ((task)->fs)
#define server_task_eq(svr,task) ((svr)->ds_proc_ident == task_pid_cookie(task))

/*	task_pid_cookie() and server_task_eq() are meant to abstract
	how we recognize threads within a single process.  Since
	Linux (kernel 2.1.79, at least) gives each thread a seperate
	task_struct, with different pids, but pointing to the same file
	descriptor tables, etc.

	Using the fs_struct to identify a process was selected, since
	at the time of do_door_shutdown (when the "/dev/door" file
	descriptor is being closed), the task's files_struct will be NULL.
*/

#define DOOR_ARGBUF_NUM_DOORS (DOOR_ARGBUF_SIZE / sizeof(struct door*))

typedef union door_argbuf {
	struct list_head	da_free;	/* link on door_argbufs */
	char			da_buffer[DOOR_ARGBUF_SIZE];
	struct door		*da_doors[DOOR_ARGBUF_NUM_DOORS];
} door_argbuf_t;

struct door_message {
	struct list_head	dm_msgs;	/* link on dr_msgs */
	struct task_struct	*dm_client;
	struct task_struct	*dm_server;
	door_arg_t		*dm_params;
	struct wait_queue	*dm_wait;
	int			dm_status;
	int			dm_error;
	struct list_head	dm_inprogress;	/* link on ds_inprogress */
	door_argbuf_t		*dm_arg_buf;
	door_argbuf_t		*dm_door_buf;
};
#define DM_IN_PROGRESS		0x0000
#define DM_DONE			0x0001

struct door {
	struct door_server	*dr_server;
	struct list_head	dr_msgs;
	struct list_head	dr_doors;	/* link on ds_doors */
	struct wait_queue	*dr_wait_avail_servers;
	void			*dr_server_proc;
	void			*dr_cookie;
	door_attr_t		dr_attributes;
	int			dr_count;
};

/* In order to have the correct metadata semantics (e.g. chmod(),
 * etc.) in the implementation of fattach, there's a many-to-one
 * mapping from door inodes to struct door, so there's no
 * door_to_inode() operation.  We need to maintain independent owner,
 * permissions, etc., for each file mount point, so each one gets a
 * seperate inode.
 */

struct door_inode_info {
	struct door		*di_door;
	struct dentry		*di_covers;
};

/* NOTE:  get_door_inode_info() is kind of sick, but we know that the
 * inode info union (inode.u) is larger than two words, however I'd be
 * very careful when adding to struct door_inode_info for fear of
 * overrunning the inode's size.  This was done for speed and
 * simplicity, otherwise we'd either have to modify Linux's struct
 * inode or kmalloc() a door_inode_info and use inode.u.generic_ip.
 */

#define get_door_inode_info(ino) ((struct door_inode_info*) &(ino)->u)

#define inode_to_door(ino) (get_door_inode_info(ino)->di_door)
#define door_valid(door) ((door)->dr_server != NULL)
#define door_mkinvalid(door) do { (door)->dr_server = NULL; } while(0)
#define door_uniquifier(door) ((door_id_t)(door))

static int door_wake_server(struct door_server *svr, struct door *dp);

/* Module global data
 * ------------------
 */

static struct super_operations door_sbops = {
	NULL,		/* read_inode */
	NULL,		/* write_inode */
	door_put_inode,	/* put_inode */
	door_delete_inode,	/* delete_inode */
	NULL,		/* notify_change */
	NULL,		/* put_super */
	NULL,		/* write_super */
	NULL,		/* statfs */
	NULL		/* remount_fs */
};

static struct super_block door_sb = {
	0,		/* s_dev */
	0,		/* s_blocksize */
	0,		/* s_blocksize_bits */
	0,		/* s_lock */
	0,		/* s_rd_only */
	0,		/* s_dirt */
	NULL,		/* s_type */
	&door_sbops,	/* s_op */
	/* ... etc, all zero ... */
};

static struct file_operations door_fops = {
	NULL,		/* lseek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* poll */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	door_open,	/* open */
	door_release,	/* release */
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL		/* lock */
};

static struct inode_operations door_inode_ops = {
	&door_fops,	/* default_file_ops */
	NULL,		/* create */
	NULL,		/* lookup */
	NULL,		/* link */
	NULL,		/* unlink */
	NULL,		/* symlink */
	NULL,		/* mkdir */
	NULL,		/* rmdir */
	NULL,		/* mknod */
	NULL,		/* rename */
	NULL,		/* readlink */
	NULL,		/* follow_link */
	NULL,		/* readpage */
	NULL,		/* writepage */
	NULL,		/* bmap */
	NULL,		/* truncate */
	NULL,		/* permission */
	NULL,		/* smap */
	NULL,		/* updatepage */
	NULL		/* revalidate */
};

static struct list_head door_servers;
static struct list_head door_argbufs;
static struct list_head door_messages;

/* ------------------ */

#if DOORS_DEBUG_MODULE

static const char *doors_assert_fmt
	= KERN_ERR "%s: Assertion failure in %s, line %d.\n";

#if DEBUG_REFCOUNTS

static inline void door_dentry_counts(const char *label, struct dentry *dent)
{
	struct inode *ino;
	struct door *dr;

	ino = dent->d_inode;
	dr = inode_to_door(ino);

	printk("%s: counts: dent %ld, ino %ld, door %ld\n",
		label,
		(unsigned long) dent->d_count,
		(unsigned long) ino->i_count,
		(unsigned long) dr->dr_count);
}

#endif /* DEBUG_MODULE && DEBUG_REFCOUNTS */
#endif /* DEBUG_MODULE */

int door_init(void)
{
#if DOORS_DEBUG_MODULE
	printk("%s: installed module.\n", doors_banner);
#endif
	INIT_LIST_HEAD(&door_servers);
	INIT_LIST_HEAD(&door_argbufs);
	INIT_LIST_HEAD(&door_messages);
	return 0;
}

static inline struct door_message *door_message_alloc(void)
{
	struct door_message *msg;

	if (list_empty(&door_messages)) {
		dprintk(DEBUG_ALLOC, ("door_message_alloc: allocating "
			"new door_message\n"));
		msg = (struct door_message*)
			kmalloc(sizeof(struct door_message), GFP_KERNEL);
	}
	else {
		msg = list_entry(door_messages.next,
			struct door_message, dm_msgs);
		list_del(&msg->dm_msgs);
	}

	return msg;
}

static inline void door_message_free(struct door_message *msg)
{
	list_add(&msg->dm_msgs, &door_messages);
}

static inline door_argbuf_t *door_argbuf_alloc(void)
{
	door_argbuf_t *ab;

	if (list_empty(&door_argbufs)) {
		dprintk(DEBUG_ALLOC, ("door_argbuf_alloc: allocating buffer "
			"of size %d\n", DOOR_ARGBUF_SIZE));
		ab = (door_argbuf_t*)
			kmalloc(sizeof(door_argbuf_t), GFP_KERNEL);
	}
	else {
		ab = list_entry(door_argbufs.next, door_argbuf_t, da_free);
		list_del(&ab->da_free);
	}

#if DOORS_DEBUG_MODULE && DOOR_ARGBUF_POISON
	memset(ab, 0xCC, sizeof(*ab));
#endif
	return ab;
}

static inline void door_argbuf_free(door_argbuf_t *ab)
{
#if DOORS_DEBUG_MODULE && DOOR_ARGBUF_POISON
	memset(ab, 0xDD, sizeof(*ab));
#endif
	list_add(&ab->da_free, &door_argbufs);
}

static inline void init_door_message(struct door_message *msg,
	door_arg_t *params, struct door *dp)
{
	msg->dm_params = params;
	init_waitqueue(&msg->dm_wait);
	msg->dm_status = DM_IN_PROGRESS;
	msg->dm_error = 0;
	msg->dm_client = current;
	msg->dm_server = NULL;
	msg->dm_arg_buf = NULL;
	msg->dm_door_buf = NULL;
	list_add(&msg->dm_msgs, &dp->dr_msgs);
}

static inline void door_message_free_buffers(struct door_message *msg)
{
	if (msg->dm_arg_buf) {
		door_argbuf_free(msg->dm_arg_buf);
		msg->dm_arg_buf = NULL;
	}

	if (msg->dm_door_buf) {
		door_argbuf_free(msg->dm_door_buf);
		msg->dm_door_buf = NULL;
	}
}

static inline int is_door_inode(struct inode *ino)
{
	return (ino->i_op == &door_inode_ops);
}

static struct door *door_get(int fd)
{
	struct file *filp;
	struct dentry *dent;
	struct inode *ino;
	struct door *dr;

	filp = fget(fd);
	if (!filp)
		goto bad_fd;
	dent = filp->f_dentry;
	if (!dent)
		goto bad_dent;
	ino = dent->d_inode;
	if (!ino || !is_door_inode(ino))
		goto bad_ino;

	dr = inode_to_door(ino);
	if (!door_valid(dr))
		goto bad_ino;

	++dr->dr_count;

	DOOR_DENTRY_COUNTS("door_get", dent);

	fput(filp);
	return dr;

bad_ino:
bad_dent:
	fput(filp);
bad_fd:
	return NULL;
}

static inline void door_free(struct door *dr)
{
	dprintk(DEBUG_ALLOC, ("door_free: freeing door %lx\n",
		(unsigned long) door_uniquifier(dr)));

	if (dr->dr_server != NULL)
		list_del(&dr->dr_doors);

	kfree(dr);
}

static inline void door_send_unref(struct door *dr)
{
	if (!(dr->dr_attributes & DOOR_UNREF))
		return;

	if ((dr->dr_attributes & DOOR_UNREF_PENDING) ||
		(dr->dr_attributes & DOOR_UNREF_DELIVERED))
		return;

	if (!dr->dr_server)
		return;

	dr->dr_attributes |= DOOR_UNREF_PENDING;
	door_wake_server(dr->dr_server, dr);
}

static inline void door_put(struct door *dr)
{
	dprintk(DEBUG_REFCOUNTS, ("door_put: door count %ld\n",
		(unsigned long) dr->dr_count - 1));

	switch (--dr->dr_count) {
	case 1:
		door_send_unref(dr);
		break;
	case 0:
		door_free(dr);
		break;
	}
}

static inline struct door_server *find_server(struct task_struct *task)
{
	struct list_head *tmp;
	struct door_server *svr;

	for (tmp = door_servers.next; tmp != &door_servers; tmp = tmp->next) {
		svr = list_entry(tmp, struct door_server, ds_servers);
		if (server_task_eq(svr, task))
			return svr;
	}
	return NULL;
}

static struct door_server *door_server_alloc(void)
{
	struct door_server *svr;

	svr = kmalloc(sizeof(struct door_server), GFP_KERNEL);
	if (!svr)
		goto bad_kmalloc;

	INIT_LIST_HEAD(&svr->ds_doors);
	svr->ds_pid = current->pid;
	svr->ds_proc_ident = task_pid_cookie(current);
	init_waitqueue(&svr->ds_wait_svr_create);
	init_waitqueue(&svr->ds_wait_avail_servers);
	init_waitqueue(&svr->ds_wait_svr_create_ready);
	init_waitqueue(&svr->ds_wait_server_ready);
	svr->ds_needs_svr_create = 0;
	INIT_LIST_HEAD(&svr->ds_inprogress);

	list_add(&svr->ds_servers, &door_servers);

	dprintk(DEBUG_ALLOC, ("door_server_alloc: allocated server %lx; "
		"pid %d\n",
		(unsigned long) svr,
		(int) svr->ds_pid));

	return svr;

bad_kmalloc:
	return NULL;
}

static inline void door_server_free(struct door_server *svr)
{
	struct list_head *dr_list;
	struct list_head *dr_it;
	struct door *dr;

	dprintk(DEBUG_ALLOC, ("door_server_free: freeing server %lx\n",
		(unsigned long) svr));

	list_del(&svr->ds_servers);

	dr_list = &svr->ds_doors;

	for (dr_it = dr_list->next; dr_it != dr_list; dr_it = dr_it->next) {
		dr = list_entry(dr_it, struct door, dr_doors);
		list_del(&dr->dr_doors);
		door_mkinvalid(dr);
	}

	/* XXX Wake processes waiting on the server and make them
	   return error codes to their callers.
	*/

	kfree(svr);
}

static inline struct door_server *find_server_alloc(struct task_struct *task)
{
	struct door_server *svr;

	if ((svr = find_server(task)) == NULL)
		svr = door_server_alloc();

	return svr;
}

static struct inode *door_inode_alloc(struct door *dr, struct inode *inherit)
{
	struct inode *ino;
	struct door_inode_info *info;

	ino = get_empty_inode();
	if (!ino)
		return 0;

	if (inherit) {
		ino->i_mode = S_IFIFO | (inherit->i_mode & S_IRWXUGO);
		ino->i_uid = inherit->i_uid;
		ino->i_gid = inherit->i_gid;
		ino->i_atime = inherit->i_atime;
		ino->i_mtime = inherit->i_mtime;
		ino->i_ctime = inherit->i_ctime;
	}
	else {
		ino->i_mode = S_IFIFO | DOOR_DEFAULT_MODE;
		ino->i_uid = current->uid;
		ino->i_gid = current->gid;
		ino->i_atime = ino->i_mtime = ino->i_ctime = CURRENT_TIME;
	}

	ino->i_nlink = 0;
	ino->i_sock = 0;
	ino->i_op = &door_inode_ops;
	ino->i_sb = &door_sb;

	info = get_door_inode_info(ino);
	info->di_door = dr;
	info->di_covers = NULL;

	return ino;
}

static inline int may_fattach(struct inode *mnt, struct inode *cover)
{
	/* XXX Does Solaris use these permissions?
	*/
	if (S_ISDIR(cover->i_mode) || S_ISDIR(mnt->i_mode))
		return -EINVAL;
	if (suser())
		return 0;
	if (current->uid != cover->i_uid)
		return -EPERM;
	return permission(cover, MAY_WRITE);
}

int do_door_fattach(int fd, const char *name)
{
	struct file *filp;
	struct dentry *cover;
	struct dentry *mount;
	struct inode *ino;
	int error;

	lock_kernel();

	filp = fget(fd);
	if (!filp) {
		error = -EBADF;
		goto bad_file;
	}

	/* Is the input fd a door?
	*/
	if (!is_door_inode(filp->f_dentry->d_inode)) {
		error = -EINVAL;
		goto bad_door_inode;
	}

	cover = lookup_dentry(name, NULL, 1);
	if (IS_ERR(cover)) {
		error = PTR_ERR(cover);
		goto bad_lookup_dentry;
	}

	if (!cover->d_inode) {
		error = -ENOENT;
		goto out;
	}

	/* Is the file to cover a mount point for a filesystem or for
	 * a door?
	 */
	if (cover->d_covers != cover || is_door_inode(cover->d_inode)) {
		error = -EBUSY;
		goto out;
	}

	error = may_fattach(filp->f_dentry->d_inode, cover->d_inode);
	if (error)
		goto out;

	ino = door_inode_alloc(inode_to_door(filp->f_dentry->d_inode),
		cover->d_inode);
	if (!ino)
		goto out;

	/* Now we proceed to remove the mount point from the dcache
	 * and replace it with the door dentry.  Note that we don't
	 * dput(cover), so that is stays alive.
	 */

	mount = d_alloc(cover->d_parent, &cover->d_name);
	if (!mount) {
		error = -ENOMEM;
		goto bad_d_alloc;
	}

	if (!cover->d_inode || list_empty(&cover->d_hash)) {
		/* While d_alloc was sleeping, the covered directory
		 * entry was unlinked.
		 */
		error = -ENOENT;
		dput(mount);
		goto out;
	}

	ino->i_nlink = 1;

	d_drop(cover);
	d_add(mount, ino);

	/* Associate the new inode with the covered dentry.
	*/
	get_door_inode_info(ino)->di_covers = cover;

	/* The fattach() counts as a reference to the door.
	*/
	inode_to_door(ino)->dr_count++;

	DOOR_DENTRY_COUNTS("door_fattach", mount);

	cover = NULL;	/* Don't drop the reference count on the dentry. */
	error = 0;

out:
bad_d_alloc:
	if (cover)
		dput(cover);
bad_door_inode:
bad_lookup_dentry:
	dprintk(DEBUG_REFCOUNTS, ("door_fattach: fput fd %d, count %ld\n",
		(int) fd,
		(unsigned long) filp->f_count - 1));
	fput(filp);
bad_file:
	unlock_kernel();
	return error;
}

int do_door_fdetach(const char *name)
{
	struct dentry *mount;
	struct dentry *cover;
	int error;

	lock_kernel();

	mount = lookup_dentry(name, NULL, 1);
	if (IS_ERR(mount)) {
		error = PTR_ERR(mount);
		mount = NULL;
		goto out;
	}

	if (!mount->d_inode) {
		error = -ENOENT;
		goto out;
	}

	if (!is_door_inode(mount->d_inode)) {
		error = -EINVAL;
		goto out;
	}

	cover = get_door_inode_info(mount->d_inode)->di_covers;
	if (!cover) {
		printk(KERN_WARNING "door_fdetach: door file with no covered "
			"dentry???\n");
		error = -EINVAL;
		goto out;
	}

	error = may_fattach(mount->d_inode, cover->d_inode);
	if (error)
		goto out;

	ASSERT(mount->d_inode->i_nlink == 1);

	mount->d_inode->i_nlink = 0;

	/* Now remove the door file from the dcache.  The next lookup
	 * will hit the original file.
	 */
	if (!list_empty(&mount->d_hash))
		d_drop(mount);

	/* We put mount twice, since one reference is held from the
	 * above lookup_dentry() and one was the extra reference from
	 * do_door_fattach() to keep it in the dcache.
	 */
	dput(mount);

	DOOR_DENTRY_COUNTS("door_fdetach", mount);

	error = 0;

out:
	if (mount)
		dput(mount);
	unlock_kernel();
	return error;
}

static struct door *door_alloc(void)
{
	struct door *dr;
	struct door_server *svr;

	dr = (struct door*) kmalloc(sizeof(struct door), GFP_KERNEL);
	if (!dr)
		goto bad_door_alloc;

	dr->dr_server_proc = NULL;
	dr->dr_cookie = NULL;
	dr->dr_attributes = 0;
	dr->dr_count = 0;
	INIT_LIST_HEAD(&dr->dr_msgs);

	svr = find_server_alloc(current);
	if (!svr)
		goto bad_svr_alloc;
	list_add(&dr->dr_doors, &svr->ds_doors);
	dr->dr_server = svr;

	return dr;

bad_svr_alloc:
	kfree(dr);
bad_door_alloc:
	return NULL;
}

static int door_open(struct inode *ino, struct file *filp)
{
	struct door *dr;

	dr = inode_to_door(ino);
	if (!dr || !is_door_inode(ino)) {
		printk(KERN_ERR "%s: bad door inode in door_open!\n",
			doors_banner);
		return -EINVAL;
	}

	dr->dr_count++;

	dprintk(DEBUG_DOOR_OPENCLOSE || DEBUG_REFCOUNTS,
		("door_open: %lx, count = %ld\n",
		(unsigned long) dr,
		(unsigned long) dr->dr_count));

	return 0;
}

static int door_release(struct inode *ino, struct file *filp)
{
	struct door *dr;

	dr = inode_to_door(ino);

	dprintk(DEBUG_DOOR_OPENCLOSE || DEBUG_REFCOUNTS,
		("door_release: %lx, count = %ld\n",
		(unsigned long) dr,
		(unsigned long) dr->dr_count - 1));

	door_put(dr);

	return 0;
}

static void door_put_inode(struct inode *ino)
{
	dprintk(DEBUG_REFCOUNTS, ("door_put_inode: count %ld, nlink %ld\n",
		(unsigned long) ino->i_count - 1,
		(unsigned long) ino->i_nlink));
}

static void door_delete_inode(struct inode *ino)
{
	struct dentry *covers;
	struct door *dr;

	dr = inode_to_door(ino);

	dprintk(DEBUG_ALLOC, ("door_delete_inode: %lx\n",
		dr));

	covers = get_door_inode_info(ino)->di_covers;
	if (covers) {
		dprintk(DEBUG_REFCOUNTS, ("door_delete_inode: "
			"d_count cover %ld\n",
			(unsigned long) covers->d_count - 1));

		get_door_inode_info(ino)->di_covers = NULL;
		dput(covers);

		/* fattach()'d inodes contribute to the reference
		 * count on the door.
		 */
		door_put(dr);
	}
}

static int door_create_fd(struct door *dr)
{
	int fd;
	struct file *filp;
	struct inode *ino;
	int error;

	fd = get_unused_fd();
	if (fd < 0)
		return fd;

	error = -ENFILE;
	filp = get_empty_filp();
	if (!filp)
		goto bad_filp;

	filp->f_op = &door_fops;
	filp->f_mode = 3;
	filp->f_flags = O_RDWR;
	filp->f_pos = 0;

	error = -ENOMEM;
	ino = door_inode_alloc(dr, NULL);
	if (!ino)
		goto bad_inode_alloc;

	error = -ENOMEM;
	filp->f_dentry = d_alloc_root(ino, NULL);
	if (!filp->f_dentry)
		goto bad_dent;

	current->files->fd[fd] = filp;
	dr->dr_count++;

	DOOR_DENTRY_COUNTS("door_create_fd", filp->f_dentry);

	return fd;

bad_dent:
bad_inode_alloc:
	put_filp(filp);
bad_filp:
	put_unused_fd(fd);
	return error;
}

int do_door_create(void *server_procedure, void *cookie, u_int attributes)
{
	int error;
	struct door *dr;

	lock_kernel();

	error = -EINVAL;
	if (attributes & ~NP_DOOR_ATTRS)
		goto bad_attributes;

	error = -ENFILE;
	dr = door_alloc();
	if (!dr)
		goto bad_door_alloc;

	dr->dr_server_proc = server_procedure;
	dr->dr_cookie = cookie;
	dr->dr_attributes = attributes;

	error = door_create_fd(dr);
	if (error < 0)
		goto bad_create_fd;

	unlock_kernel();
	return error;

bad_create_fd:
	door_free(dr);
bad_door_alloc:
bad_attributes:
	unlock_kernel();
	return error;
}

int do_door_shutdown(void)
{
	struct door_server *svr;

	if ((svr = find_server(current)) != NULL)
		door_server_free(svr);

	return 0;
}

int do_door_wait_svr_create(door_info_t *info)
{
	struct door_server *svr;

	svr = find_server_alloc(current);
	if (!svr)
		goto bad_svr_alloc;

	while (!svr->ds_needs_svr_create) {
		if (signal_pending(current))
			return -ERESTARTSYS;
		wake_up(&svr->ds_wait_svr_create_ready);
		interruptible_sleep_on(&svr->ds_wait_svr_create);
	}

	svr->ds_needs_svr_create = 0;

	return 0;

bad_svr_alloc:
	return -ENOMEM;
}

static int door_create_server_thread(struct door_server *svr, struct door *dp)
{
	while (!waitqueue_active(&svr->ds_wait_svr_create)) {
		if (signal_pending(current))
			return -EINTR;
		interruptible_sleep_on(&svr->ds_wait_svr_create_ready);

		/* While we were waiting for the "server create"
		 * thread to become available, a running server thread
		 * might have returned.
		 */
		if (waitqueue_active(&svr->ds_wait_avail_servers))
			return 0;
	}

	if (signal_pending(current))
		return -EINTR;

	svr->ds_needs_svr_create = 1;

	wake_up(&svr->ds_wait_svr_create);
	interruptible_sleep_on(&svr->ds_wait_server_ready);

	return 0;
}

static int door_wake_server(struct door_server *svr, struct door *dp)
{
	int error;

	while (!waitqueue_active(&svr->ds_wait_avail_servers)) {
		error = door_create_server_thread(svr, dp);
		if (error)
			return error;
	}

	/* XXX: Waking up all servers, even though only one will
	 * handle this request.
	 */
	wake_up(&svr->ds_wait_avail_servers);

	return 0;
}

static inline int door_copyout_one_door(door_desc_t *udesc, struct door *dr)
{
	int fd;
	door_attr_t attr;
	door_desc_t desc;

	dprintk(DEBUG_DOOR_ARGCHECK,
		("door_copyout_doors: about to create fd for door %lx\n",
		(unsigned long) dr));

	fd = door_create_fd(dr);
	if (fd < 0)
		return fd;

	attr = (dr->dr_attributes & DOOR_UATTR_MASK) | DOOR_DESCRIPTOR;

	/* Was this door created by the receiving process?
	*/
	if (server_task_eq(dr->dr_server, current))
		attr |= DOOR_LOCAL;

	/* The caller should have called verify_area().
	   We assume that the user memory is valid for write operations.
	*/

	dprintk(DEBUG_DOOR_ARGCHECK, ("door_copyout_doors: writing "
		"door_desc_t at %lx in user space\n",
		(unsigned long) udesc));

	desc.d_attributes = attr;
	desc.d_data.d_desc.d_descriptor = fd;
	desc.d_data.d_desc.d_id = door_uniquifier(dr);

	if (copy_to_user(udesc, &desc, sizeof(desc)))
		return -EFAULT;

	dprintk(DEBUG_TRANSLATE, ("door_copyout_doors: sent door %lx "
		"to fd %d\n",
		(unsigned long) door_uniquifier(dr),
		(int) fd));
	return 0;
}

static int door_copyout_doors(door_desc_t *udesc, struct door **kdoors,
	size_t desc_num)
{
	size_t snork;
	int error;

	/* Must be called within the context of the process receiving
	   the doors.
	*/

	for (snork = 0; snork < desc_num; snork++, kdoors++) {
		error = door_copyout_one_door(udesc++, *kdoors);
		if (error)
			goto err_out;
	}

	return 0;

err_out:
	/* XXX clean up all translated doors on error
	*/
	return error;
}

static inline int door_copyin_doors(struct door **kdoors,
	door_desc_t *udoors, size_t num_doors)
{
	int fd;
	int dd;
	int error;
	struct door *dr;
	door_desc_t desc;

	for (dd = 0; dd < num_doors; dd++, udoors++, kdoors++) {
		error = -EFAULT;
		if (copy_from_user(&desc, udoors, sizeof(desc)))
			goto cleanup;

		/* Only passing doors supported for now.
		*/
		error = -EINVAL;
		if (!(desc.d_attributes & DOOR_DESCRIPTOR))
			goto cleanup;

		fd = desc.d_data.d_desc.d_descriptor;

		error = -EBADF;
		dr = door_get(fd);
		if (!dr)
			goto cleanup;

		*kdoors = dr;

		dprintk(DEBUG_TRANSLATE, ("door_copyin_doors: received "
			"door %lx from fd %d\n",
			(unsigned long) door_uniquifier(dr),
			(int) fd));
	}

	return 0;

cleanup:
	dprintk(DEBUG_TRANSLATE, ("door_copyin_doors: an error occurred while "
		"translating doors; cleaning up\n"));
	while (dd--)
		door_put(*(--kdoors));
	return error;
}

static inline size_t door_buffer_size(size_t data_size, size_t desc_num)
{
	return (DOOR_DESC_ALIGN(data_size) + desc_num * sizeof(door_desc_t));
}

static inline int door_call_copy_inargs(struct door_message *msg)
{
	dprintk(DEBUG_CALL, ("door_call: copying inargs, data_size %ld, "
		"desc_num %ld\n",
		(unsigned long) msg->dm_params->data_size,
		(unsigned long) msg->dm_params->desc_num));

	if (msg->dm_params->data_size &&
		msg->dm_params->data_size <= DOOR_ARGBUF_SIZE) {
		door_argbuf_t *ab;

		ab = door_argbuf_alloc();
		if (!ab)
			return -EAGAIN;

		if (copy_from_user(ab->da_buffer, msg->dm_params->data_ptr,
			msg->dm_params->data_size)) {
			door_argbuf_free(ab);
			return -EFAULT;
		}

		msg->dm_arg_buf = ab;
	}

	if (msg->dm_params->desc_num) {
		door_argbuf_t *ab;
		int error;

		/* Arbitrary limit on the number of door descriptors
		   which can be passed in a single door_call.
		*/
		if (msg->dm_params->desc_num > DOOR_ARGBUF_NUM_DOORS)
			return -EINVAL;

		ab = door_argbuf_alloc();
		if (!ab)
			return -EAGAIN;

		error = door_copyin_doors(ab->da_doors,
			msg->dm_params->desc_ptr, msg->dm_params->desc_num);
		if (error) {
			door_argbuf_free(ab);
			return error;
		}

		msg->dm_door_buf = ab;
	}

	/* Otherwise, we'll map data into the server from within
	   door_return_copy_inargs. */

	return 0;
}

static inline int door_call_copy_outargs(struct door_message *msg)
{
	dprintk(DEBUG_CALL, ("door_call: copying outargs\n"));

	/* All of the fields of the door_arg_t will have been filled
	   in by door_return_copy_outargs.
	*/

	if (msg->dm_arg_buf) {
		if (copy_to_user(msg->dm_params->data_ptr,
			msg->dm_arg_buf->da_buffer,
			msg->dm_params->data_size))
			return -EFAULT;
	}

	if (msg->dm_door_buf) {
		int error;

		error = door_copyout_doors(msg->dm_params->desc_ptr,
			msg->dm_door_buf->da_doors, msg->dm_params->desc_num);
		if (error)
			return error;	/* XXX clean up */
	}

	return 0;
}

int do_door_call(int fd, door_arg_t *args)
{
	struct door *dp;
	int error;
	struct door_message *msg;
	struct door_server *svr;

	lock_kernel();

	dp = door_get(fd);
	if (!dp) {
		error = -EBADF;
		goto bad_file;
	}

	svr = dp->dr_server;

	error = -ENOMEM;
	msg = door_message_alloc();
	if (!msg)
		goto out_nomem;
	init_door_message(msg, args, dp);

	error = door_call_copy_inargs(msg);
	if (error)
		goto remove_call;

	error = door_wake_server(svr, dp);
	if (error)
		goto remove_call;

	while (msg->dm_status != DM_DONE) {
		if (signal_pending(current)) {
			error = -EINTR;
			goto cancel_call;
		}
		interruptible_sleep_on(&msg->dm_wait);
	}

	error = msg->dm_error;
	if (!error)
		error = door_call_copy_outargs(msg);

	dprintk(DEBUG_CALL, ("door_call: returning from call\n"
		"   data_ptr %lx, data_size %ld,\n"
		"   desc_ptr %lx, desc_num %ld,\n"
		"   rbuf %lx, rsize %ld\n",
		(unsigned long) msg->dm_params->data_ptr,
		(unsigned long) msg->dm_params->data_size,
		(unsigned long) msg->dm_params->desc_ptr,
		(unsigned long) msg->dm_params->desc_num,
		(unsigned long) msg->dm_params->rbuf,
		(unsigned long) msg->dm_params->rsize));

	door_message_free_buffers(msg);
	door_message_free(msg);

out_nomem:
	door_put(dp);
bad_file:
	unlock_kernel();
	return error;

cancel_call:
	msg->dm_client = NULL;
	if (!msg->dm_server) {
		door_message_free_buffers(msg);
		door_message_free(msg);
	}
	door_put(dp);
	unlock_kernel();
	return error;

remove_call:
	list_del(&msg->dm_msgs);
	door_message_free_buffers(msg);
	msg->dm_client = NULL;
	door_put(dp);
	unlock_kernel();
	return error;
}

static inline int get_message(struct door_server *svr,
	struct door **dr_ret, struct door_message **msg_ret)
{
	struct list_head *dr_it;
	struct list_head *dr_list = &svr->ds_doors;
	struct door *dr;

	for (dr_it = dr_list->next; dr_it != dr_list; dr_it = dr_it->next) {
		dr = list_entry(dr_it, struct door, dr_doors);
		if (!list_empty(&dr->dr_msgs)) {
			struct list_head *msg_it = dr->dr_msgs.next;
			struct door_message *msg;

			msg = list_entry(msg_it, struct door_message, dm_msgs);
			list_del(&msg->dm_msgs);

			*dr_ret = dr;
			*msg_ret = msg;
			return 1;
		}
	}
	return 0;
}

static inline struct door_message *get_inprogress(struct door_server *svr)
{
	struct list_head *msg_it;
	struct list_head *msg_list = &svr->ds_inprogress;
	struct door_message *msg;

	for (msg_it = msg_list->next; msg_it != msg_list;
		msg_it = msg_it->next) {
		msg = list_entry(msg_it, struct door_message, dm_inprogress);
		if (msg->dm_server == current) {
			list_del(&msg->dm_inprogress);
			return msg;
		}
	}
	return NULL;
}

static inline int door_return_copy_inargs(struct door_message *msg,
	char *argstash, size_t argstash_size,
	struct door_return_outargs *outargs)
{
	unsigned long udesc_addr;

	/* Some trickery here, since the input data can either go
	   in a static buffer (aka. argstash) or in the newly mapped
	   segment.  We'll delay translating the doors until we have
	   the memory in the user's address space.  Note that the
	   doors always go into the argstash.
	*/

	outargs->munmap_addr = NULL;
	udesc_addr = (unsigned long) argstash;

	/* The user library must make sure that argstash_size
	   is greater than or equal to DOOR_ARGBUF_SIZE, or else
	   the server is toast.
	*/

	dprintk(DEBUG_RETURN, ("door_return: copying inargs, %s copy args, "
		"%s door args, argstash %lx\n",
		msg->dm_arg_buf ? "has" : "no",
		msg->dm_door_buf ? "has" : "no",
		(unsigned long) argstash));

	if (msg->dm_arg_buf) {
		if (copy_to_user(argstash, msg->dm_arg_buf->da_buffer,
			msg->dm_params->data_size))
			return -EFAULT;

		outargs->data_ptr = argstash;
		outargs->data_size = msg->dm_params->data_size;

		udesc_addr = (unsigned long) (argstash +
			msg->dm_params->data_size);
		udesc_addr = DOOR_DESC_ALIGN(udesc_addr);
	}
	else if (msg->dm_params->data_size) {
		unsigned long addr;
		int error;

		/* XXX Perhaps there should be a limit on the amount
		   of data the client can allocate in the server?
		*/
		error = door_vm_copy(current, msg->dm_client,
			(unsigned long) msg->dm_params->data_ptr,
			(unsigned long) msg->dm_params->data_size, 0,
			&addr);
		if (error)
			return error;

		outargs->data_ptr = (void*) addr;
		outargs->data_size = msg->dm_params->data_size;
		outargs->munmap_addr = (void*) (addr & PAGE_MASK);
		outargs->munmap_len = PAGE_ALIGN(msg->dm_params->data_size +
			(addr & ~PAGE_MASK));
	}
	else {
		outargs->data_ptr = NULL;
		outargs->data_size = 0;
	}

	if (msg->dm_door_buf) {
		int error;

		error = door_copyout_doors((door_desc_t*) udesc_addr,
			msg->dm_door_buf->da_doors, msg->dm_params->desc_num);
		if (error)
			return error;

		dprintk(DEBUG_RETURN, ("door_return: %ld doors translated, "
			"no errors\n",
			(unsigned long) msg->dm_params->desc_num));

		outargs->desc_ptr = (door_desc_t*) udesc_addr;
		outargs->desc_num = msg->dm_params->desc_num;
	}
	else {
		outargs->desc_ptr = NULL;
		outargs->desc_num = 0;
	}

	return 0;
}

static inline int door_return_copy_outargs(struct door_message *msg,
	void *data_ptr, size_t data_size,
	door_desc_t *desc_ptr, size_t desc_num)
{
	size_t buffer_size;

	/* Note that the average case door_return() may or may not
	 * contain data, but probably doesn't contain descriptors.
	 * This is slightly optimized for that case.
	 */

	msg->dm_params->data_size = data_size;
	msg->dm_params->desc_num = desc_num;

	if (!data_size)
		msg->dm_params->data_ptr = NULL;
	if (!desc_num)
		msg->dm_params->desc_ptr = NULL;

	if (!data_size && !desc_num)
		return 0;

	buffer_size = door_buffer_size(data_size, desc_num);

	dprintk(DEBUG_RETURN, ("door_return: outargs require a buffer size of "
		"%ld bytes\n", (unsigned long) buffer_size));

	if (buffer_size <= msg->dm_params->rsize) {
		if (data_size) {
			msg->dm_arg_buf = door_argbuf_alloc();
			if (!msg->dm_arg_buf)
				return -EAGAIN;

			if (copy_from_user(msg->dm_arg_buf->da_buffer,
				data_ptr, data_size))
				return -EFAULT;

			msg->dm_params->data_ptr = msg->dm_params->rbuf;
		}
	}
	else {
		unsigned long addr;
		int error;
		unsigned long alloc_len;

		/* If we have descriptors to pass back, and the result buffer
		   isn't big enough, we need to allocate extra memory in the
		   process address space.
		*/
		if (desc_num) {
			alloc_len = DOOR_DESC_ALIGN(data_size)
				- data_size
				+ sizeof(door_desc_t) * desc_num;
			dprintk(DEBUG_RETURN, ("door_return: "
				"adding %ld bytes to vm allocation "
				"for %ld door descriptors\n",
				(unsigned long) alloc_len,
				(unsigned long) desc_num));
		}
		else
			alloc_len = 0;

		/* If there's no data, door_vm_copy will only allocate
		   memory for the descriptors.
		*/
		if (!data_size)
			data_ptr = NULL;

		error = door_vm_copy(msg->dm_client, current,
			(unsigned long) data_ptr,
			(unsigned long) data_size,
			alloc_len,
			&addr);
		if (error)
			return error;

		if (data_size)
			msg->dm_params->data_ptr = (void*) addr;

		msg->dm_params->rbuf = (void*) (addr & PAGE_MASK);
		msg->dm_params->rsize = data_size + alloc_len;
	}

	if (desc_num) {
		int error;
		door_argbuf_t *ab;

		/* Arbitrary limit on the number of door descriptors
		   which can be passed in a single door_call.
		*/
		if (desc_num > DOOR_ARGBUF_NUM_DOORS)
			return -EINVAL;

		ab = door_argbuf_alloc();
		if (!ab)
			return -ENOMEM;

		error = door_copyin_doors(ab->da_doors, desc_ptr, desc_num);
		if (error) {
			door_argbuf_free(ab);
			return error;
		}

		msg->dm_door_buf = ab;

		if (data_size)
			msg->dm_params->desc_ptr = (void*)
				DOOR_DESC_ALIGN((unsigned long)
					msg->dm_params->data_ptr +
					msg->dm_params->data_size);
		else
			msg->dm_params->desc_ptr = (void*)
				DOOR_DESC_ALIGN((unsigned long)
					msg->dm_params->rbuf);
	}

	return 0;
}

static int get_unref(struct door_server *svr, struct door **dr_ret)
{
	struct list_head *dr_it;
	struct list_head *dr_list = &svr->ds_doors;
	struct door *dr;

	/* Only send unref messages to doors which don't have any
	 * messages remaining to be processed.
	 */

	for (dr_it = dr_list->next; dr_it != dr_list; dr_it = dr_it->next) {
		dr = list_entry(dr_it, struct door, dr_doors);
		if ((dr->dr_attributes & DOOR_UNREF_PENDING) &&
			list_empty(&dr->dr_msgs)) {
			*dr_ret = dr;
			return 1;
		}
	}
	return 0;
}

static int door_return_unref(struct door *dr,
	struct door_return_outargs *outargs)
{
	/* This door has received its one and only DOOR_UNREF_DATA
	 * message and it no longer needs to be sent.
	 */
	dr->dr_attributes |= DOOR_UNREF_DELIVERED;
	dr->dr_attributes &= ~DOOR_UNREF_PENDING;

	outargs->proc = dr->dr_server_proc;
	outargs->cookie = dr->dr_cookie;
	outargs->data_ptr = DOOR_UNREF_DATA;
	outargs->data_size = 0;
	outargs->desc_ptr = 0;
	outargs->desc_num = 0;
	outargs->munmap_addr = 0;
	outargs->munmap_len = 0;

	return 0;
}

int do_door_return(void *data_ptr, size_t data_size, door_desc_t *desc_ptr,
	size_t desc_num, char *argstash, size_t argstash_size,
	unsigned long munmap_addr, unsigned long munmap_len,
	struct door_return_outargs *outargs)
{
	struct door_server *svr;
	struct door *dr;
	struct door_message *msg;
	int error;

	lock_kernel();

	/* do_door_return() has two purposes.  The first part of the
	 * implementation returns arguments to the invoking client
	 * thread.  The second part of the implementation waits for
	 * the next request that can be handled by the current server
	 * thread.
	 */

	svr = find_server(current);
	if (!svr) {
		printk(KERN_ERR "%s: no door_server in door_return?!?!?\n",
			doors_banner);
		send_sig(SIGKILL, current, 0);
		error = -EINVAL;
		goto out;
	}

	msg = get_inprogress(svr);
	if (msg) {
		dprintk(DEBUG_RETURN, ("door_return: responding to message; "
			"data_ptr %lx, data_size %ld, "
			"desc_ptr %lx, desc_num %ld\n",
			(unsigned long) data_ptr,
			(unsigned long) data_size,
			(unsigned long) desc_ptr,
			(unsigned long) desc_num));

		if (msg->dm_client) {
			error = door_return_copy_outargs(msg, data_ptr,
				data_size, desc_ptr, desc_num);

			msg->dm_status = DM_DONE;
			msg->dm_error = error;
			msg->dm_server = NULL;
			wake_up(&msg->dm_wait);
		}
		else {
			dprintk(DEBUG_RETURN, ("door_return: client aborted; "
				"ignoring message response\n"));
			door_message_free_buffers(msg);
			door_message_free(msg);
		}
	}
	msg = NULL;

	/* After we've responded to a message in progress, if any,
	   it's safe to unmap the address region which we might have
	   mapped on behalf of the server process.
	*/
	if (munmap_addr) {
		dprintk(DEBUG_INARGS_UNMAP, ("door_return: unmapping %lx "
			"-> %lx (size %ld)\n",
			(unsigned long) munmap_addr,
			(unsigned long) munmap_addr + munmap_len,
			(unsigned long) munmap_len));

		if (do_munmap(munmap_addr, munmap_len))
			printk(KERN_WARNING "door_return: error unmapping "
				"overflow region!\n");
	}

	/* At this point, we've done all of the necessary processing
	 * to return values to the invoking client thread.  Now we
	 * wait for the next message that this server thread can
	 * handle.
	 */

retry:
	for (;;) {
		msg = NULL;

		if (get_unref(svr, &dr)) {
			/* If there's an error sending the DOOR_UNREF
			 * message, there's really nothing more we can
			 * do besides continue processing messages.
			 */
			error = door_return_unref(dr, outargs);
			if (!error)
				goto out;
		}

		if (get_message(svr, &dr, &msg))
			break;

		error = -ERESTARTSYS;
		if (signal_pending(current))
			goto out;
		wake_up(&svr->ds_wait_server_ready);
		interruptible_sleep_on(&svr->ds_wait_avail_servers);
	}

	if (!msg->dm_client) {
		dprintk(DEBUG_RETURN, ("door_return: client aborted; "
			"ignoring message\n"));
		door_message_free_buffers(msg);
		door_message_free(msg);
		goto retry;
	}

	msg->dm_server = current;
	list_add(&msg->dm_inprogress, &svr->ds_inprogress);

	outargs->proc = dr->dr_server_proc;
	outargs->cookie = dr->dr_cookie;

	outargs->cred.dc_euid = msg->dm_client->euid;
	outargs->cred.dc_ruid = msg->dm_client->uid;
	outargs->cred.dc_egid = msg->dm_client->egid;
	outargs->cred.dc_rgid = msg->dm_client->gid;
	outargs->cred.dc_pid = msg->dm_client->pid;

	error = door_return_copy_inargs(msg, argstash, argstash_size,
		outargs);
	if (error) {
		dprintk(DEBUG_RETURN, ("door_return: error %d copying inargs, "
			"waiting for next request\n", error));

		msg->dm_status = DM_DONE;
		msg->dm_error = error;
		wake_up(&msg->dm_wait);
		goto retry;
	}

	/* Done with the temporary message passing buffers.
	   They can be reused while the call is being handled.
	*/
	door_message_free_buffers(msg);

	dprintk(DEBUG_RETURN, ("door_return: returning with request, "
		"server_proc %lx, cookie %lx, data_ptr %lx, data_size %ld, "
		"desc_ptr %lx, desc_num %ld\n",
		(unsigned long) outargs->proc,
		(unsigned long) outargs->cookie,
		(unsigned long) outargs->data_ptr,
		(unsigned long) outargs->data_size,
		(unsigned long) outargs->desc_ptr,
		(unsigned long) outargs->desc_num));

	error = 0;

out:
	unlock_kernel();
	return error;
}

int do_door_info(int fd, struct door_info *info)
{
	struct door *dr;
	struct door_server *svr;

	dr = door_get(fd);
	if (!dr)
		return -EBADF;

	svr = dr->dr_server;

	info->di_target = svr->ds_pid;
	info->di_proc = dr->dr_server_proc;
	info->di_data = dr->dr_cookie;
	info->di_attributes = dr->dr_attributes & DOOR_UATTR_MASK;
	info->di_uniquifier = door_uniquifier(dr);

	/* Was this door created by the receiving process?
	*/
	if (server_task_eq(dr->dr_server, current))
		info->di_attributes |= DOOR_LOCAL;

	door_put(dr);
	return 0;
}
