/* Doors for Linux.
   Copyright (C) 1998 Jason Lango

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
**  test_large.c - Send/receive large arguments.
**  Jason Lango <jal@cs.brown.edu>
*/

#include "test.h"

#define DATA_LEN (2 * 1024 * 1024)

void init_data(char *data, size_t len)
{
	int ii;
	long *lp;

	lp = (long*) data;
	for (ii = 0; ii < (len / sizeof(long)); ii++)
		*lp++ = ii;
}

int verify_data(char *data, size_t len)
{
	int ii;
	long *lp;

	lp = (long*) data;
	for (ii = 0; ii < (len / sizeof(long)); ii++)
		if (*lp++ != ii)
			return 0;

	return 1;
}

void sigcleanup(int signo)
{
	test_destroy_door();
	exit(0);
}

void server_proc(void *cookie, char *argp, size_t arg_size,
	door_desc_t *dp, size_t n_desc)
{
	door_cred_t cred;
	static char *data = NULL;

	if (cookie != TEST_COOKIE) {
		test_error("server_proc: cookie %lx != %lx\n",
			(long) cookie, (long) TEST_COOKIE);
	}

	if (!argp || arg_size != DATA_LEN) {
		test_error("server_proc: expecting arguments; got %lx %lx\n",
			(unsigned long) argp, (unsigned long) arg_size);
	}

	if (dp || n_desc)
		test_error("server_proc: received unexpected doors\n");

	VERIFY ( door_cred(&cred) );

	if (dts->dts_verbose) {
		printf("server_proc(cookie=0x%08lx)\n", (long) cookie);
		printf("  client pid=%d, uid=%d, gid=%d, euid=%d, egid=%d\n",
			cred.dc_pid,
			cred.dc_ruid, cred.dc_rgid,
			cred.dc_euid, cred.dc_egid);
	}

	test_verify_cred(&cred);

	if (!verify_data(argp, arg_size))
		test_error("server: argument is garbled\n");

	/* Modify the input argument, to prove that it's copy-on-write.
	*/
	memset(argp, 0xCC, arg_size);

	/* Return data can't be modified by receiver.
	*/
	if (data && !verify_data(data, DATA_LEN))
		test_error("server: return data modified\n");

	if (!data)
		data = malloc(DATA_LEN);
	if (!data)
		test_error("server_proc: data malloc failed!\n");
	init_data(data, DATA_LEN);

	VERIFY ( door_return(data, DATA_LEN, NULL, 0) );
	test_error("door_return() returned\n");
}

const char *test_description(void)
{
	return "large argument passing";
}

int server_main(int num, int argc, char *argv[])
{
	int door;

	signal(SIGINT, sigcleanup);
	signal(SIGTERM, sigcleanup);
	signal(SIGHUP, sigcleanup);

	door = test_create_door(server_proc);

	for (;;)
		pause();

	return 0;
}

int client_main(int num, int argc, char *argv[])
{
	int fd;
	int ntries;
	int max_tries = 5;
	door_arg_t args;
	char *data;

	if (argc)
		max_tries = strtol(argv[0], NULL, 10);

	if (dts->dts_verbose) {
		printf("%s: calling %d times...\n", dts->dts_testname,
			max_tries);
	}

	data = malloc(DATA_LEN);
	init_data(data, DATA_LEN);

	fd = test_open_door();

	for (ntries = 0; ntries < max_tries; ntries++) {
		args.data_ptr = data;
		args.data_size = DATA_LEN;
		args.desc_ptr = NULL;
		args.desc_num = 0;
		args.rbuf = NULL;
		args.rsize = 0;

		VERIFY ( door_call(fd, &args) );

		if (!verify_data(data, DATA_LEN))
			test_error("client: door_call() changed data\n");

		if (!args.rbuf || args.rsize < DATA_LEN)
			test_error("door_call: bad result buffer\n");

		if (!args.data_ptr || args.data_size != DATA_LEN)
			test_error("client: door_call: invalid data\n");

		if (args.data_ptr == data) {
			test_error("client: door_call: "
				"data_ptr didn't change\n");
		}

		if (!verify_data(args.data_ptr, args.data_size))
			test_error("client: returned data garbled\n");

		if (args.desc_ptr || args.desc_num)
			test_error("door_call: returned unexpected doors\n");

		/* Modify result data, to prove that it's copy-on-write.
		*/
		memset(args.data_ptr, 0xDD, args.data_size);

		if (dts->dts_verbose)
			printf("%s: client: unmapping result buffer\n",
				dts->dts_testname);
		VERIFY ( munmap(args.rbuf, args.rsize) );
	}

	free(data);

	test_kill_servers();
	return 0;
}
