/* Doors for Linux.
   Copyright (C) 1998 Jason Lango

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
**  test_pass.c - Test door descriptor passing.
**  Jason Lango <jal@cs.brown.edu>
*/

#include "test.h"

int return_this_door = -1;
int num_calls = 0;

const char *test_description(void)
{
	return "door descriptor passing";
}

void sigcleanup(int signo)
{
	test_destroy_door();

	if ((num_calls % 2) != 0)
		test_error("client didn't make 2 calls\n");

	exit(0);
}

int info_equal(struct door_info *a, struct door_info *b)
{
	if (a->di_uniquifier != b->di_uniquifier ||
		a->di_target != b->di_target || a->di_proc != b->di_proc ||
		a->di_data != b->di_data)
		return 0;
	else
		return 1;
}

void server_proc(void *cookie, char *argp, size_t arg_size,
	door_desc_t *dp, size_t n_desc)
{
	door_cred_t cred;
	door_desc_t desc;
	int door;
	struct door_info info;
	struct door_info other_info;

	++num_calls;

	if (cookie != TEST_COOKIE) {
		test_error("cookie %lx != %lx\n", (unsigned long) cookie,
			(unsigned long) TEST_COOKIE);
	}

	VERIFY ( door_cred(&cred) );

	test_verify_cred(&cred);

	if (argp || arg_size) 
		test_error("server_proc received unexpected arguments\n");

	if (!dp || !n_desc)
		test_error("server_proc expecting doors\n");

	if (!(dp->d_attributes & DOOR_DESCRIPTOR))
		test_error("d_attributes without DOOR_DESCRIPTOR\n");

	door = dp->d_data.d_desc.d_descriptor;

	VERIFY ( door_info(door, &info) );
	test_verify_info(&info, server_proc);

	VERIFY ( door_info(return_this_door, &other_info) );
	test_verify_info(&other_info, server_proc);

	/* This door should be the one we've created in server_main().
	*/
	if (!info_equal(&info, &other_info))
		test_error("server: sent/received doors don't match\n");

	VERIFY ( close(door) );

	desc.d_attributes = DOOR_DESCRIPTOR;
	desc.d_data.d_desc.d_descriptor = return_this_door;

	VERIFY ( door_return(NULL, 0, &desc, 1) );
	test_error("door_return shouldn't return");
}

int server_main(int num, int argc, char *argv[])
{
	signal(SIGINT, sigcleanup);
	signal(SIGTERM, sigcleanup);
	signal(SIGHUP, sigcleanup);

	return_this_door = test_create_door(server_proc);

	for (;;)
		pause();

	test_error("server_main shouldn't return\n");
	return 1;
}

/* do_pass() sends this door through itself (whoa!) and then expects
 * to receive a new door descriptor to the same door back from the
 * server.
 */

int do_pass(int door)
{
	door_arg_t args;
	door_desc_t pass_me;
	char res[256];
	int other_door;
	struct door_info info, other_info;

	pass_me.d_attributes = DOOR_DESCRIPTOR;
	pass_me.d_data.d_desc.d_descriptor = door;

	args.data_ptr = NULL;
	args.data_size = 0;
	args.desc_ptr = &pass_me;
	args.desc_num = 1;
	args.rbuf = res;
	args.rsize = sizeof(res);

	VERIFY ( door_call(door, &args) );

	if (args.data_ptr || args.data_size)
		test_error("client: received unexpected arguments\n");

	if (!args.desc_ptr || !args.desc_num)
		test_error("client: expecting returned doors\n");

	if (args.rbuf != res || args.rsize != sizeof(res))
		test_error("client: result buffer moved unnecessarily\n");

	if (!(args.desc_ptr->d_attributes & DOOR_DESCRIPTOR)) {
		test_error("client: d_attributes %lx without DOOR_DESCRIPTOR\n",
			(unsigned long) args.desc_ptr->d_attributes);
	}

	other_door = args.desc_ptr->d_data.d_desc.d_descriptor;

	if (other_door < 0 || other_door == door) {
		test_error("client: returned door %d invalid\n",
			other_door);
	}

	VERIFY ( door_info(door, &info) );
	test_verify_info(&info, server_proc);

	VERIFY ( door_info(other_door, &other_info) );
	test_verify_info(&other_info, server_proc);

	if (!info_equal(&info, &other_info))
		test_error("client: sent/received doors don't match\n");

	return other_door;
}

int client_main(int num, int argc, char *argv[])
{
	int door;
	int ii;
	int ntries;
	int recv_door, recv_door2;

	ntries = 100;
	if (argc)
		ntries = strtol(argv[0], NULL, 10);

	door = test_open_door();

	for (ii = 0; ii < ntries; ii++)
	{
		recv_door = do_pass(door);
		recv_door2 = do_pass(recv_door);
		VERIFY ( close(recv_door) );
		VERIFY ( close(recv_door2) );
	}

	test_kill_servers();
	return 0;
}
