/* Doors for Linux.
   Copyright (C) 1998 Jason Lango

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
**  test_small.c - Send/receive small arguments.
**  Jason Lango <jal@cs.brown.edu>
*/

#include "test.h"

const char	CALL_STR[]	= "Hello, World!";
#define		CALL_LEN	(sizeof(CALL_STR))

const char	RETURN_STR[]	= "How's it going?";
#define		RETURN_LEN	(sizeof(RETURN_STR))

void sigcleanup(int signo)
{
	test_destroy_door();
	exit(0);
}

void server_proc(void *cookie, char *argp, size_t arg_size,
	door_desc_t *dp, size_t n_desc)
{
	door_cred_t cred;

	if (cookie != TEST_COOKIE)
		test_error("server_proc: cookie %lx != %lx\n",
			(long) cookie, (long) TEST_COOKIE);

	if (!argp || !arg_size)
		test_error("server_proc: expecting arguments; got %lx %lx\n",
			(unsigned long) argp, (unsigned long) arg_size);

	if (dp || n_desc)
		test_error("server_proc: received unexpected doors\n");

	VERIFY ( door_cred(&cred) );

	if (dts->dts_verbose) {
		printf("server_proc(cookie=0x%08lx)\n", (long) cookie);
		printf("  client pid=%d, uid=%d, gid=%d, euid=%d, egid=%d\n",
			cred.dc_pid,
			cred.dc_ruid, cred.dc_rgid,
			cred.dc_euid, cred.dc_egid);
		printf("  argument: \"%s\"\n", argp);
		printf("  replying: \"%s\"\n", RETURN_STR);
	}

	test_verify_cred(&cred);

	if (arg_size != CALL_LEN)
		test_error("server_proc: invalid arg_size %ld\n",
			(long) arg_size);

	if (!argp || memcmp(CALL_STR, argp, CALL_LEN))
		test_error("server_proc: invalid argument \"%s\"\n",
			argp);

	VERIFY ( door_return((char*)RETURN_STR, RETURN_LEN, NULL, 0) );
	test_error("door_return() returned\n");
}

const char *test_description(void)
{
	return "small argument passing";
}

int server_main(int num, int argc, char *argv[])
{
	int door;

	signal(SIGINT, sigcleanup);
	signal(SIGTERM, sigcleanup);
	signal(SIGHUP, sigcleanup);

	door = test_create_door(server_proc);

	for (;;)
		pause();

	return 0;
}

int client_main(int num, int argc, char *argv[])
{
	int fd;
	int ntries;
	int max_tries = 20;
	door_arg_t args;
	char res[255];

	if (argc)
		max_tries = strtol(argv[0], NULL, 10);

	if (dts->dts_verbose) {
		printf("%s: calling %d times...\n", dts->dts_testname,
			max_tries);
	}

	fd = test_open_door();

	for (ntries = 0; ntries < max_tries; ntries++) {
		memset(res, 0, sizeof(res));

		args.data_ptr = (char*) CALL_STR;
		args.data_size = CALL_LEN;
		args.desc_ptr = NULL;
		args.desc_num = 0;
		args.rbuf = res;
		args.rsize = sizeof(res);

		VERIFY ( door_call(fd, &args) );

		if (args.rbuf != res || args.rsize != sizeof(res))
			test_error("door_call: relocated result buffer\n");

		if (!args.data_ptr || args.data_size != RETURN_LEN)
			test_error("door_call: returned invalid data\n");

		if (args.desc_ptr || args.desc_num)
			test_error("door_call: returned unexpected doors\n");

		if (dts->dts_verbose)
			printf("sent \"%s\", received \"%s\"\n",
				CALL_STR, args.data_ptr);
	}

	test_kill_servers();
	return 0;
}
